#!/usr/bin/perl
#
# $Author: swaj $
# $Revision: 1.16 $
#
# Simple PTk Database.
#
# Copyright (c) Alexey Semenoff 2001-2007. All rights reserved.
# Distributed under GNU Public License.
#
#

use 5.008;
use strict;
use warnings;
use Encode qw /encode decode from_to/;
use Data::Dumper;
use Getopt::Long;
use DBI;

use Tk;
use Tk::Dialog;
use Tk::HList;
use Tk::ItemStyle;

use vars qw /
    $toplevel
    $toplevel_sp
    $toplevel_nt
    $toplevel_nr
    $toplevel_sb

    $entry_text

    $mw
    $menu
    $ff
    $ff2a
    $ff2
    $ff3
    $fh
    $fm

    $fr3
    $fb
    $lab
    $lab2

    $opt_tables
    $opt

    $list
    $entry

    $dbh
    $sth

    $current_table
    $prev_h

    %encode_converts
    $encode_convert
    $prev_h
    $start_from
    %config
    $PACKAGE
    $OS
    $debug
    $NAME

    $file

    $logo
    $img
    $VERSION

    $sqlite_version

    @sqlite_tables

    $new_table_name
    $new_table_columns

    @INSERT_RECORD
    @INSERT_RECORD_LAST
    @CUR_COLUMNS

    $id_for_insert

    $sort_order
    $order_by
    $found_total
    /;


warn 'Tested with Tk 804.027 only!' if ($Tk::VERSION < 804.027);


BEGIN {
    $PACKAGE = __PACKAGE__;

    if ( $^O =~ /^(ms)?(dos|win(32|nt)?)/i ) {
        $OS = 'WIN';
    } else {
        $OS = 'UNIX';
    }

    $_=$0;
    s|^(.+)/.*|$1|;
    $start_from = $_;

    $debug = 0;
    $NAME = 'Data';
    $prev_h = 0;

    $toplevel_sp = q{};
    $toplevel_nt = q{};
    $toplevel_nr = q{};

    @sqlite_tables = ();
    $current_table = q{};
    $sort_order = 1;
    $found_total = 0;
    $VERSION = '1.2.3';
};

sub printd (;@);
sub printe (;@);
sub printi (;@);

use constant PROGRAM_HOME_DIR		=> '/opt/sdict';

use constant DATA_CONFIG_FILE		=> '.ptksdictrc';
use constant DATA_CONFIG_FILE_DEF	=> 'ptksdictrc.sample';
use constant DATA_CONFIG_FILE_WIN	=> 'ptksdictrc.win';
use constant DATA_FOUND_TOTAL		=> 'Found: ';

binmode STDOUT, ':utf8';

#####
# S #
#####


GetOptions("debug" => \$debug);

ParseConfig();

DrawInterface();

if ( ($OS eq 'UNIX') && ! $config{ 'disable_switcher' } && (exists ($config{'default_encoding'})) )
{

    if (exists ($encode_converts{$config{'default_encoding'}})) {
	$encode_convert = $config{'default_encoding'};
	printi "switch to '$encode_convert' encoding";
	ChangeEncoding();
    }
}

#printd "> $config{'database'}";


$dbh = DBI->connect("dbi:SQLite:dbname=$config{'database'}",
			"",
			"",
			{
				RaiseError => 0,
				AutoCommit => 1 #,
	    			#HandleError => sub { printe 'ERROR'; }
 			}
 		    );

unless ($dbh) {
  my $err = "Unable connect to database, make sure '$config{'database'}' is valid";
  printe $err;
  DisplayError($err);
  exit 1;
}

$sqlite_version = $dbh->{sqlite_version}; 

ScanTables();

$opt_tables->addOptions (@sqlite_tables);

MainLoop;

####
#  #
####


sub printd (;@) {
    $debug && print "DEBUG ($PACKAGE): ", @_, "\n";
}


sub printe (;@) {
    print "ERROR ($PACKAGE): ", @_, "\n";
}


sub printi (;@) {
    print "INFO  ($PACKAGE): ", @_, "\n";
}


sub NewTable {
    printd "NewTable()";
    return if $toplevel_nt;

    $new_table_name = $new_table_columns = q{};


    $toplevel_nt = $mw->Toplevel();
    $toplevel_nt->title('Create new table');


    my $fr0 = $toplevel_nt->Frame ()->pack(-side =>'top',-fill => 'x');

    $fr0->Label (
		 -text => 'Table name: ',
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'left',
						 -expand => 'yes',
						 -padx => 1,
						 -pady => 1,
						 -fill => 'both'
						 );
    $fr0->Entry(
		-width => 50,
		-font => $config{'font'},
		-textvariable => \$new_table_name
		)-> pack(
			 -side =>'left',
			 -fill => 'x'
			 );


    my $fr1 = $toplevel_nt->Frame ()->pack(-side =>'top',-fill => 'x');


    $fr1->Label (
		 -text => 'Columns (space separated): ',
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'left',
						 -expand => 'yes',
						 -padx => 1,
						 -pady => 1,
						 -fill => 'both'
						 );
    $fr1->Entry(
		-width => 50,
		-font => $config{'font'},
		-textvariable => \$new_table_columns
		)-> pack(
			 -side =>'right',
			 -fill => 'x',
			 -padx => 1,
			 -pady => 1
			 );


    my $fr2 = $toplevel_nt->Frame () -> pack(
					      -side => 'bottom'
					      #,-fill => 'x'
					      );
    $fr2->Button ( -text => 'Create',
		   -borderwidth => 1,
		   -command => [ \&OkCreateNewTable ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'left');

    $fr2->Button ( -text => 'Cancel',
		   -borderwidth => 1,
		   -command => [ \&CloseCreateNewTable ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'right');

    CenterWindow($toplevel_nt);

}


sub CloseCreateNewTable {
    printd "CloseCreateNewTable()";
    $toplevel_nt->destroy;
    $toplevel_nt = q{};
}


sub OkCreateNewTable {
    my $err;
    printd "OkCreateNewTable()";
    return unless $toplevel_nt;

    my @columns = split /\s+/, $new_table_columns;

    if ( ($new_table_columns eq q{}) ||
	 ($new_table_name eq q{}) ||
	 (! scalar(@columns)) ||
	 ($new_table_columns =~ /^\s+$/) ||
	 ($new_table_name =~ /^\s+$/)
	 ) {

	DisplayError ('Wrong parameters');
	return;
    }


    unshift @columns, 'id INTEGER';

    my $ll = join ',', @columns;
    my $oper = "CREATE TABLE $new_table_name ( $ll)"; 

    printd ">$oper<";

    unless ($dbh->do($oper)) {
	$err = 	"SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    push @sqlite_tables, $new_table_name;
    $opt_tables->addOptions ($new_table_name);

    CloseCreateNewTable();
}


sub DropTable {
    my $err;
    printd "DropTable()";
    return if $toplevel_nt;

    my $button = $mw->messageBox('-icon' => 'info', -type => 'YesNo',
	-title => 'Drop table',
        -message => 'Drop current table?');
    printd "selected '$button'";
    return if ($button ne 'Yes');

    printd "dropping table '$current_table'";

    return unless $fm;
    return if ($current_table eq q{});


    my $oper = "DROP TABLE $current_table"; 

    printd ">$oper<";

    unless ($dbh->do($oper)) {
	$err = 	"SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    CreateDefaultTableFrame();

    @sqlite_tables = grep(!/^$current_table$/, @sqlite_tables);
    $current_table = $sqlite_tables[0];

#    $opt_tables->configure ( -options => @sqlite_tables);
    $opt_tables->configure ( -options => [ () ] );

    $opt_tables->addOptions (@sqlite_tables);

    ChangeTable();
}


sub EditRecord {
    my $err;
    printd "EditRecord()";

    printd "table = $current_table";

    return if ($current_table eq q{});

    my $id = $list->infoAnchor;

    return if (!defined $id || !$id);

    InsertRecord_v($id);
}


sub InsertRecord {
    InsertRecord_v(0);
}


sub InsertRecord_v {
    my $err;
    my $id = shift;
    $id_for_insert = $id;

    printd "InsertRecord() // id = $id";

    printd "table = $current_table";

    $sth = $dbh->prepare("SELECT sql FROM sqlite_master WHERE name = '$current_table'");

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    my @row = $sth->fetchrow_array;
    $sth->finish;


    my $a = $row[0];
    chomp ($a);
    $a =~ s|^.+?\((.+?)\).*$|$1|;
    $a =~ s|\s||g;
	
    printd ">$a<";

    my @b = split /,/, $a;

    shift @b;

    my @roww = ();

    if ($id) {

	$sth = $dbh->prepare("SELECT * FROM $current_table WHERE id = $id");

	unless ($sth) {
	    $err = "SQL prepare failed: " . $dbh->errstr;
	    DisplayError ($err);
	    return;
	}

	unless ($sth->execute()) {
	    $err = "SQL execute failed: " . $dbh->errstr;
	    DisplayError ($err);
	    return;
	}

	@roww = $sth->fetchrow_array;
	$sth->finish;
    }


#    printd @roww ; die;

#    @INSERT_RECORD
    return if $toplevel_nr;

    $toplevel_nr = $mw->Toplevel();
    $toplevel_nr->title('Insert data');

    @INSERT_RECORD = ();
    @INSERT_RECORD_LAST = ();

    $INSERT_RECORD[0] = $current_table;

    my $ii = 1;

    for my $nn (@b) {

	my $fr0 = $toplevel_nr->Frame () -> pack (
						  -side   => 'top'  ,
						  -fill   => 'both' ,
						  -expand => 'yes'  ,
						 );

	$fr0->Label (
		     -text => "$nn: ",
		     -font => $config{'font_bold'},
		     -justify => 'center' ) -> pack (
						     -side   => 'left' ,
						     -expand => 'no'   ,
						     -padx   => 1      ,
						     -pady   => 1      ,
						     -fill   => 'both' ,
						     );

	my $txt = $fr0->Scrolled(
				 qw/Text -scrollbars e -wrap word width 40 heigh 2 /
				 )->pack(
					 -fill   => 'both' ,
					 -expand => 'yes'  ,
					 );



	$txt->bind (
		    '<Key>' => [ \&RecodeText, $txt, $ii ]
		    );


	if ($id) {
	    $_ = $roww[$ii];
	    $_ = decode("utf8", $_);
	    $txt->insert('end', $_);
	}


	$INSERT_RECORD_LAST[$ii] = 0;

	$INSERT_RECORD[$ii++] = $txt;
    }



    my $fr2 = $toplevel_nr->Frame () -> pack(
					      -side => 'bottom'
					      #,-fill => 'x'
					      );
    $fr2->Button ( -text => 'Save',
		   -borderwidth => 1,
		   -command => [ \&OkSaveNewRecord ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'left');

    $fr2->Button ( -text => 'Cancel',
		   -borderwidth => 1,
		   -command => [ \&CloseNewRecord ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'right');

    CenterWindow($toplevel_nr);
}


sub CloseNewRecord {
    printd "CloseNewRecord()";
    $toplevel_nr->destroy;
    $toplevel_nr = q{};
    ChangeTable();
}


sub GetNextId {
    my $err;
    my $table = shift;

    printd "table =  $table";

    $sth = $dbh->prepare("SELECT MAX (id) FROM $table");

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return 0;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return 0;
    }

    my @row = $sth->fetchrow_array;
    $sth->finish;

    my $id = 1;

#    print "row -> $row[0]";

    if (!defined ($row[0]) || !$row[0]) {
	#$id = 1;
    } else {
	$id = $row[0] + 1;

    }

    printd "new id = $id";

    return $id;
}


sub OkSaveNewRecord {
    my $err;
    printd "OkSaveNewRecord()";
    return unless $toplevel_nr;

#    print Dumper @INSERT_RECORD;

    my $table = $INSERT_RECORD[0];

    printd "table =  $table";


    my $id = GetNextId($table);
    return if ($id <=0);

    printd "new id = $id";


    my $r = q{};

    for (my $i=1; $i <= $#INSERT_RECORD; $i++) {

	my $e = $INSERT_RECORD[$i];
	my $t = $e->get('1.0', 'end');
	$t = encode("utf8", $t);
	
	$e->delete('1.0', 'end');
	chomp ($t);

	$t = decode("utf8", $t);

	printd ">>$t<<";

	$r .= ", '$t'"
    }


    unless ($id_for_insert) {

	my $stmt = "INSERT INTO  $table VALUES ('$id' $r)";
	printd "$stmt";

	unless ($dbh->do($stmt)) {
	    $err = 	"SQL execute failed: " . $dbh->errstr;
	    DisplayError ($err);
	    return;
	}

    } else {

	unless ($dbh->do("DELETE FROM $table WHERE id = $id_for_insert")) {

	    $err = 	"SQL execute failed: " . $dbh->errstr;
	    DisplayError ($err);
	    CloseNewRecord();
	    return;
	}
	
	my $stmt = "INSERT INTO  $table VALUES ('$id_for_insert' $r)";
#	printd "$stmt";

	unless ($dbh->do($stmt)) {
	    $err = 	"SQL execute failed: " . $dbh->errstr;
	    DisplayError ($err);
	    CloseNewRecord();
	    return;
	}

	CloseNewRecord();
    }

}


sub RecodeText {
    printd "RecodeTextPress()";

#    return;

    if ($OS eq 'WIN' || $config{ 'disable_switcher' } ) {
#	$update = 1;
    	#UpdateCurList();
	return;
    }

    my($widget) = shift;
    my $iidd = shift;
    my $e = $widget->XEvent;    # get event object
    my($keysym_text, $keysym_decimal) = ($e->K, $e->N);
    printd "pressed: '$keysym_text' dec: '$keysym_decimal' ";

#    if ($keysym_text eq 'Return') { printd "RET:"; $text->openLine; return; }

    my $last = $widget->get('end - 2 chars' , 'end' );
    chomp ($last); 

    printd "last = '$last'";

#    $last = encode("utf8", $last);

    my $h = ord ($last);
    
    printd "ord = '$h'";
   
    if ($h < 128 || $h > 255) {

	$prev_h = $h; #$INSERT_RECORD_LAST[$iidd] = $h;
	return;
    }

    my $res = $widget->search(-backwards => "$last", 'end');
    printd "res = '$res'";
    $widget->delete ( $res , 'end' );

#    printd "recode from: $encode_converts{$encode_convert}";

    from_to ($last, $encode_converts{$encode_convert}, "utf8");

    $last = decode("utf8", $last);

#    if ($INSERT_RECORD_LAST[$iidd] == 10) {
    if ($prev_h  == 10) {
	$widget->insert ('end', "\n");
    }

    $widget->insert ('end', $last);

    $prev_h = 0;
#    $INSERT_RECORD_LAST[$iidd] = 0;

}


sub DeleteRecord {
    my $err;
    printd "DeleteRecord()";

    my $id = $list->infoAnchor;

    return if (!defined $id || !$id);

    my $button = $mw->messageBox('-icon' => 'info', -type => 'YesNo',
	-title => 'Delete selected record',
        -message => 'Delete selected record?');
    printd "selected '$button'";
    return if ($button ne 'Yes');

    $list->selectionClear($id);

    my $stmt = "DELETE FROM $current_table WHERE id = $id";
    printd "$stmt";

    unless ($dbh->do($stmt)) {
	$err = 	"SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    ChangeTable();
}


sub ScanTables {
    my $err;
    my @tables = ();

    printd "ScanTables()";

    $sth = $dbh->prepare('SELECT name FROM sqlite_master ORDER BY 1');

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    while (my @row = $sth->fetchrow_array ) {
	printd "found table '$row[0]'";
	push @tables, $row[0];
    }

    $sth->finish;

    for my $t (@tables) {

	$sth = $dbh->prepare("SELECT sql FROM sqlite_master WHERE name = '$t'");

	unless ($sth) {
	    $err = "SQL prepare failed: " . $dbh->errstr;
	    DisplayError ($err);
	    return;
	}

	unless ($sth->execute()) {
	    $err = "SQL execute failed: " . $dbh->errstr;
	    DisplayError ($err);
	    return;
	}

	my @row = $sth->fetchrow_array;
	$sth->finish;

	my $a = $row[0];
	chomp ($a);
	$a =~ s|INTEGER||i;
	$a =~ s|^.+?\((.+?)\).*$|$1|;
	$a =~ s|\s||g;
	
	printd ">$a<";

	my @b = split /,/, $a;


	if ($b[0] !~ /^id$/i) {
	    printd "no Id column in '$t'";
	    next;
	}

	shift @b; # remoce Id

	my $l = scalar (@b);

	unless ($l) {
	    printd "no other columns in '$t'";
	    next;
	}

	push @sqlite_tables, $t;
	printd "added table '$t'";
    }
}


sub CleanQuit {
    printd "CleanQuit()";

    $dbh->disconnect if $dbh;

    $mw->destroy if $mw;
    exit 0;
}


sub SetSortBy {
    printd "SetSortBy()";
    my $err;

    return if ($current_table eq q{}|| $toplevel_sb);

    $sth = $dbh->prepare("SELECT sql FROM sqlite_master WHERE name = '$current_table'");

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    my @row = $sth->fetchrow_array;
    $sth->finish;


    my $a = $row[0];
    chomp ($a);

    $a =~ s|INTEGER||;
    $a =~ s|^.+?\((.+?)\).*$|$1|;
    $a =~ s|\s||g;
	
    printd ">$a<";

    my @b = split /,/, $a;


    $toplevel_sb = $mw->Toplevel();
    $toplevel_sb->title('Help');

    my $fr = $toplevel_sb->Frame ()->pack(-side =>'top',-fill => 'x');

    my @pl = qw/-side left -expand 1 -padx .5c -pady .5c/;

    my $left  = $fr->LabFrame(-label => 'Sort by')->pack(@pl);

    @pl = qw/-side top -pady 2 -anchor w/;


    my $i = 1;

    for my $p (@b) {

	$left->Radiobutton(
            -text     => "$p",
            -variable => \$order_by,
            -relief   => 'flat',
            -value    => $i,
        )->pack(@pl);

	$i++;
    }

    $order_by = $sort_order;

    my $fr2 = $toplevel_sb->Frame () -> pack(

					      -side => 'bottom'
					      #,-fill => 'x'
					      );
    $fr2->Button ( -text => 'Save',
		   -borderwidth => 1,
		   -command => [ \&OkApplyNewSortBy ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'left');

    $fr2->Button ( -text => 'Cancel',
		   -borderwidth => 1,
		   -command => [ \&CloseSetSortBy ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'right');

    CenterWindow($toplevel_sb);
}


sub OkApplyNewSortBy {
    my $err;
    printd "OkApplyNewSortBy()";

    return unless $toplevel_sb;

    $toplevel_sb->destroy;
    $toplevel_sb = q{};

    printd "selected new order_by = '$order_by'";
    $sort_order = $order_by;

    ChangeTable();
}


sub CloseSetSortBy {
    printd "CloseSetSortBy()";
    $toplevel_sb->destroy;
    $toplevel_sb = q{};
}


sub PrintHelp {
    printd "PrintHelp()";

    $fh->configure ( -state => 'disabled' );
    $toplevel = $mw->Toplevel();
    $toplevel->title('Help');

    my $fr = $toplevel->Frame ()->pack(-side =>'top',-fill => 'x');

    my $text = q{};

    $text = <<EOS;

PTk Data. Version: $VERSION ($OS).

Written by Alexey Semenoff [http://swaj.net] (c) 2001-2007.

Perl: $]; Tk: $Tk::VERSION; SQLite: $sqlite_version.

EOS

    $fr ->Label (
		 -image => $img
		 ) -> pack(
			   -side => 'left',
			   -padx => 15,
			   -pady => 15
			   );

    $fr->Label ( -text => $text,
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'top',
						 -expand => 'yes',
						 -padx => 10,
						 -pady => 10,
						 -fill => 'both'
						 );

    my $fr2 = $toplevel->Frame ( ) -> pack(  -side => 'bottom' , -fill => 'x');

    $fr2->Button ( -text => 'Continue',
		   -borderwidth => 1,
		   -command => [ \&ClosePrintHelp ],
		   -padx => 20,
		   -pady => 10,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'bottom');

    CenterWindow($toplevel);
}


sub ClosePrintHelp {
    $toplevel->destroy;
    $fh->configure ( -state => 'normal' );
}


sub CenterWindow {
    my $w = shift;
    return unless $w;
    $mw->update;
    $_ = $mw->geometry;
    printd "pw geometry $_";
    @_ = split /[x\+]/;
    my ($m_w, $m_h, $m_x, $m_y) = @_[0..3];
    $_ = $w->geometry;
    @_ = split /[x\+]/;
    my ($t_x, $t_y) = @_[0,1];
    my $geom = $t_x . 'x' . $t_y . '+' . int ( ($m_w  - $t_x) /2 + $m_x ) . '+' . int ( ($m_h - $t_y) / 2 + $m_y );
    $w->geometry($geom);
}


sub ExportTxt {
    my (@row, $sth, $err);

    printd "ExportTxt()";

    my $file_export = $mw->getSaveFile();

    return if (!defined($file_export)||($file_export eq q{}));

    printd "Selected file : '$file_export'\n";

    #UpdateCurList();

    unless (open (OF, "> $file_export")) {
	DisplayError("Cannot export file: $!");
	return ;
    }


    my $a = GetTableColumns ($current_table);
    my @arr = split /,/, $a;
    QuoteArray (\@arr);
    $_ = join (',', @arr);

    print OF "$_\n";


    $sth = $dbh->prepare("SELECT * FROM $current_table ORDER BY 1");

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    while (my @row = $sth->fetchrow_array ) {
	QuoteArray (\@row);
	$_ =  join (',', @row);


	if ( $config{ 'output_encoding' } )
	{
	    from_to ( $_, "utf8", $config{ 'output_encoding' } );
	}
	elsif ( ! $config{ 'disable_switcher' } && $encode_converts { $encode_convert } ne 'none' )
	{
	    from_to ($_, "utf8",  $encode_converts{$encode_convert});
	}

	print  OF "$_\n";
    }

    $sth->finish;

    close OF;
}


sub QuoteArray {
    my $arr = shift;

    for (@$arr) {
	chomp;
	s|\"||sg;
        $_ = '"' . $_ . '"';
    }
}


sub ImportTxt {
    my $err;
    printd "ImportTxt()";

    return if ($current_table eq q{});

    my $file_load = $mw->getOpenFile(); # -filetypes => \@types );

    return if (!defined($file_load)||($file_load eq q{}));

    printd "Selected file : '$file_load'\n";

    unless (open (IF, "< $file_load")) {
	DisplayError("Cannot open file: $!");
	return;
    }


    while (<IF>) {
	chomp;

	if ( $config{ 'input_encoding' } )
	{
	    from_to ( $_, $config{ 'intput_encoding' }, "utf8" );
	}
	elsif ( ! $config{ 'disable_switcher' } && $encode_converts { $encode_convert } ne 'none' )
	{
	    from_to ($_, $encode_converts{$encode_convert}, "utf8");
	}

	$_ = decode("utf8", $_);

	#printd ">$_<";

	my $id = GetNextId($current_table);


	my $oper = "INSERT INTO $current_table VALUES ($id, $_)"; 

	#printd ">$oper<"; die;

	unless ($dbh->do($oper)) {
	    $err = 	"SQL execute failed: " . $dbh->errstr;
	    DisplayError($err);
	    return;
	}

#	printd $id ; die;

    }

    close IF;

    ChangeTable($current_table);
}


sub DisplayError {
    my $text = shift;

    $text = 'Error: ' . $text;

    my $dialog = $mw->Dialog( -text => $text,
			      -font => $config{'font_bold'},
			      -bitmap => 'error',
			      -title => 'Error',
			      -default_button => 'Continue',
			      -buttons => [qw/Continue/] );
    my $d = $dialog->Show();
}


sub DisplayInfo {
    my $text = shift;

    $text = 'Info: ' . $text;

    my $dialog = $mw->Dialog( -text => $text,
			      -font => $config{'font_bold'},
			      -bitmap => 'info',
			      -title => 'info',
			      -default_button => 'Continue',
			      -buttons => [qw/Continue/] );
    my $d = $dialog->Show();
}


sub ChangeEncoding {
    printd "ChangeEncoding()";
    printd "Use encoding (main) '$encode_convert'";
}


sub DrawInterface {
    $mw = MainWindow->new;
    $mw->title($NAME);
    $mw->CmdLine;
    $mw->protocol('WM_DELETE_WINDOW' => [\&CleanQuit ]);

    $menu = $mw->Frame (
			-relief => 'raised',
			-borderwidth => 1
			) -> pack (
				   -fill =>'x'
				   );

    $logo .= $_ while (<DATA>);

    $img = $menu->Photo(
			-format => 'xpm',
			-data  => $logo
			);

    $debug || $mw->Icon(
			-image => $img
			);

    $ff = $menu->Menubutton (
			     -text => 'File',
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 1,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );
    $ff->command (
		  -label => 'New table...',
		  -font => $config{'font_bold'},
		  -command => [ \&NewTable],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Drop table',
		  -font => $config{'font_bold'},
		  -command => [ \&DropTable],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Quit',
		  -font => $config{'font_bold'},
		  -command => [ \&CleanQuit],
		  -underline => 0
		  );

#---
    $ff2a = $menu->Menubutton (
			     -text => 'View', 
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );
    $ff2a->command (
		  -label => 'Sort by...',
		  -font => $config{'font_bold'},
		  -command => [ \&SetSortBy],
		  -underline => 0
		  );

#---
    $ff2 = $menu->Menubutton (
			     -text => 'Edit', 
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );
    $ff2->command (
		  -label => 'Edit Record...',
		  -font => $config{'font_bold'},
		  -command => [ \&EditRecord],
		  -underline => 0
		  );

    $ff2->command (
		  -label => 'Insert Record...',
		  -font => $config{'font_bold'},
		  -command => [ \&InsertRecord],
		  -underline => 0
		  );

    $ff2->command (
		  -label => 'Delete Record',
		  -font => $config{'font_bold'},
		  -command => [ \&DeleteRecord],
		  -underline => 0
		  );

    $fh = $menu->Button (
			  -text => 'Help',
			  -font => $config{'font_bold'},
			  -borderwidth => 0,
			  -padx => 1,
			  -pady => 1,
			  -command => [ \&PrintHelp ],
			  -underline => 0
			  ) -> pack( -side => 'right' );

    $ff3 = $menu->Menubutton (
			     -text => 'Tools', 
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );

    $ff3->command (
		  -label => 'Export to txt...',
		  -font => $config{'font_bold'},
		  -command => [ \&ExportTxt],
		  -underline => 0
		  );


    $ff3->command (
		  -label => 'Import from txt...',
		  -font => $config{'font_bold'},
		  -command => [ \&ImportTxt],
		  -underline => 0
		  );

    $opt_tables = $menu->Optionmenu(
			       -options => [ () ] ,
			       -variable => \$current_table,
			       -command => [ \&ChangeTable ]
			       ) -> pack(
					 -side => 'left'
					 );

# Middle
    $fm = $mw->Frame (
		      -relief => 'raised',
		      -borderwidth => 1
		      ) -> pack (
				 -fill =>'both',
				  -expand => 'yes'

				 );

    CreateDefaultTableFrame();

# Bottom
    $fb = $mw->Frame (
		      -relief => 'raised',
		      -borderwidth => 1
		      ) -> pack (
				 -side => 'bottom',
				 -fill =>'x'
				 );
    
    $lab = $fb->Label (
		       -font => $config{'font_bold'},
		       -text => 'Find',
		       -justify => 'center' ) -> pack (
							-side => 'left',
							-expand => 'no',
							-padx => 1,
							-pady => 1,
							-fill => 'both'
							);
    $entry = $fb->Entry (
			 -width => 30,
			 -font => $config{'font'},
			 -textvariable => \$entry_text
			 ) -> pack (
				    -side => 'left',
				    -fill => 'x'
				    );

    $entry->bind (
		  '<Return>' => [ \&ReturnPress ]
		  );

    $entry->bind (
		  '<Key>' => [ \&EntryPress ]
		  );


    $entry->bind (
		  '<Escape>' => sub { 
			 $entry -> delete(0.1, 'end');
		     }
		  );


    if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' } ) {
	$opt = $fb->Optionmenu(
			       -options => [ () ] ,
			       -variable => \$encode_convert,
			       -command => [ \&ChangeEncoding ]
			       ) -> pack(
					 -side => 'left'
					 );
	$opt->addOptions (
			  sort keys (%encode_converts)
			  );
    }

    $lab2 = $fb->Label (
		       -font => $config{'font_bold'},
		       -text => DATA_FOUND_TOTAL . $found_total,
		       -justify => 'center' ) -> pack (
							-side => 'right',
							-expand => 'no',
							-padx => 10,
							-pady => 1,
							-fill => 'both'
						       );
    $mw->update if ($mw);
}


sub UpdateFound {
    return unless $lab2;

    $lab2->configure ( -text => DATA_FOUND_TOTAL . $found_total );
    $mw->update if ($mw);

}


sub CreateDefaultTableFrame {
    return unless $fm;
    
    $list->destroy if $list;

    $list = $fm->Scrolled ('HList',
			   -scrollbars => 'se',
			   -header => 1,
			   -indicator =>1,
			   -itemtype => 'text',
			   -columns => 2,
			   -width => 100,
			   -height => 20,
			   -selectmode => 'single',
			   -command => sub { printd ">";  printd $list->infoAnchor;   }
			   )-> pack ( #grid (
				      #qw/-sticky nsew/
				      -fill =>'both',
				      -expand => 'yes'
				      );

 
    $list->headerCreate(0, -text => "Id");
    $list->headerCreate(1, -text => "Data");

    $list->add('0');

    $list->itemCreate('0', 0, -text => '* No tables *');

    $list->configure (
		      -font => $config{'font'},
		      -columns => 2
		      );
}


sub GetTableColumns {
    my $a = q{};
    my $table = shift;
    my $err;

    return if ($table eq q{});

    $sth = $dbh->prepare("SELECT sql FROM sqlite_master WHERE name = '$table'");

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return $a;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return $a;
    }

    my @row = $sth->fetchrow_array;
    $sth->finish;

    $a = $row[0];
    chomp ($a);
    $a =~ s|INTEGER||;
    $a =~ s|^.+?\((.+?)\).*$|$1|;
    $a =~ s|\s||g;

    return $a;
    #printd ">$a<";
}


sub ChangeTable {
    my ($err, @row);
    printd "ChangeTable()";
    printd "switch to table '$current_table'";

    return unless $fm;

    if ($current_table eq q{}) {
	CreateDefaultTableFrame();
	return;
    }

    my $a = GetTableColumns ($current_table);

    my @b = split /,/, $a;

    if ($b[0] !~ /^id$/i) {
	    printd "no Id column in '$current_table'";
	    next;
	}

    @CUR_COLUMNS = @b;

    my $ll = scalar (@b);

    if ($sort_order > $ll) {
	printi '$sort_order reset';
	$sort_order = 1;
    }

    shift @CUR_COLUMNS; # remove Id

#    shift @b; # remove Id

    my $l = scalar (@b);

    printd "Len $l";
   
    my $pos = q{};
   
    $pos = $list->infoAnchor if $list;

    $list->destroy if $list;

    $list = $fm->Scrolled ('HList',
			   -scrollbars => 'se',
			   -header => 1,
			   -indicator =>1,
			   -itemtype => 'text',
			   -columns => $l,
			   -width => 100,
			   -height => 20,
			   -selectmode => 'single',
			   -command => sub { DoubleClick ($list);  }
			   )-> pack ( #grid (
				      #qw/-sticky nsew/
				      -fill =>'both',
				      -expand => 'yes'
				      );

    $list->bind (
                '<ButtonPress-3>' => [ \&RightButtonMenu, $list ]
                );


    for (my $i=0; $i<$l; $i++) {
	$list->headerCreate($i, -text => $b[$i]);
    }

#    $sort_order = 10;

    my $stmt = GetTableSelect();
    
    $sth = $dbh->prepare($stmt);

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    $found_total=0;

    while (@row = $sth->fetchrow_array) {
	$found_total++;
	my $rr = $row[0]; #shift (@row);
	$list->add($rr);

	my $ii = 0;
	for my $j (@row) {
	    $j = decode("utf8", $j);
	    $list->itemCreate($rr, $ii, -text => $j);
	    $ii++;
	}
    }

    $sth->finish;

    $list->configure (
		      -font => $config{'font'},
		      -columns => $l
		      );

    if (defined ($pos) && $pos ne q{} && $list->infoExists($pos)) {
	printd "adjust position to '$pos'";
	$list->see ($pos);
	$list->selectionSet($pos);
    }

    UpdateFound();
}


sub DoubleClick {
    printd 'DoubleClick()';
    my $lst = shift;

    printd $lst->infoAnchor;

    EditRecord();

}


sub RightButtonMenu {
    printd 'RightButtonMenu()';
    my $lst = shift;
}


sub GetTableSelect {
    printd "GetTableSelect()";

    my $et = ( defined ($entry_text) && $entry_text ne {} ) ? $entry_text : q{};
    
    my $stmt = q{};

    if ( $et eq q{} ) {

	$stmt = "SELECT * FROM $current_table ORDER BY $sort_order";

    } else {

	$stmt = "SELECT * FROM $current_table WHERE ";

	for my $j (@CUR_COLUMNS) {
	    $stmt .= "$j LIKE '%$et%' OR ";
	}	
	$stmt =~ s|OR $||;
	$stmt .= " ORDER BY $sort_order";
	
    }

    printd "got '$stmt'";
    return $stmt;
}


sub ReturnPress {
    my $err;
    printd "ReturnPress()";

    $list->delete('all', 'end');

    my $stmt = GetTableSelect();

    $sth = $dbh->prepare($stmt);

    unless ($sth) {
	$err = "SQL prepare failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    unless ($sth->execute()) {
	$err = "SQL execute failed: " . $dbh->errstr;
	DisplayError ($err);
	return;
    }

    $found_total=0;

    while (my @row = $sth->fetchrow_array) {
	$found_total++;
	my $rr = $row[0]; #shift (@row);
	$list->add($rr);

	my $ii = 0;
	for my $j (@row) {
	    $j = decode("utf8", $j);
	    $list->itemCreate($rr, $ii, -text => $j);
	    $ii++;
	}
    }

    $sth->finish;

    $list->configure (
		      -font => $config{'font'}
		      );

    UpdateFound();
}


sub EntryPress {
    my $s_enc;
    my $i=0;

    printd "EntryPress()";

    my $e = $entry->XEvent;
    my($keysym_text, $keysym_decimal) = ($e->K, $e->N);

    printd "Key '$keysym_text'";

    return if (!defined ($entry_text) || ! length($entry_text) );

    printd "entry '$entry_text'";

    my $s = substr($entry_text, 0, length($entry_text) - 1);
    my $n = substr($entry_text, length($entry_text) - 1, 1);


    if ( $keysym_text !~ /BackSpace/i )
    { 
	if ( $OS eq 'UNIX' && ! $config{ 'disable_switcher' } )
	{
	    if (unpack ("C", $n) >= 128) {
		$s_enc = $encode_converts{$encode_convert};
		
		if ( defined($s_enc) && ($s_enc ne q{}) && ($s_enc ne 'none')) {
		    from_to ($n, $s_enc, "utf8" );
		}
		
		$n = decode("utf8", $n);
		$entry_text = $s . $n;
	    }
	}
    }
}


sub ParseConfig {
    my $file;
    my %tempcfg = ();
    my @temp = ();
    my @temp2 = ();

    if ($OS eq 'WIN') {
	$file = $start_from . '/../etc/' . DATA_CONFIG_FILE_WIN;
    } else { 
	$file = "$ENV{'HOME'}/" . DATA_CONFIG_FILE;

	unless (-e $file) {
	    my $o = $file;
	    $file = PROGRAM_HOME_DIR . '/etc/' . DATA_CONFIG_FILE_DEF;
	    printi "rc file '$o' not found, using global '$file' instead";
	}
    }

    if (open (F, "< $file")) {
	while (<F>) {
	    chomp;
	    s/\r$//g;
	    next if /^\#/;
	    next if /^\s*$/;
	    my ($p,$v)=(split(/\s=\s/, $_, 2))[0,1];
	    $p =~ s|^\s+||g;
	    $p =~ s|\s+$||g;
	    $v =~ s|^\s+||g;
	    $v =~ s|\s+$||g;
	    next unless ($p && $v);
	    $tempcfg{$p} = $v;
	}
	close F;
    } else {
	printd "Unable read configuration from file '$file'";
    }

    @temp = ();
    @temp = split /\s+/, $tempcfg{'input_encodings'} if ( $tempcfg{'input_encodings'} ); 

    if ( @temp )
    {
	$config{ 'input_encodings' } = \@temp;
	$config{ 'disable_switcher' } = 0 ;
    } else {
	$config{ 'input_encodings' } = [ ] ; #  [ qw / rus koi8-r lat iso8859-1/ ];
	$config{ 'disable_switcher' } = 1 ;
    }

    $config{'font'} = $tempcfg{'font'} || '-*-*-*-*-*--*-*-*-*-*-*-iso10646-*';

    unless (exists ($tempcfg{'font_bold'})) {
	$config{'font_bold'} = $config{'font'};
	$config{'font_bold'} =~ s|^(\-.+?\-.+?)\-.+?\-(.*)$|$1-bold-$2|;
    } else {
	$config{'font_bold'} = $tempcfg{'font_bold'};
    }

    %encode_converts = @{$config{'input_encodings'}};

    $config{'database'} = $tempcfg{'database'} || "$ENV{'HOME'}/data/maindb";

    if (exists ($tempcfg{'default_encoding'})) {
	$config{'default_encoding'} = $tempcfg{'default_encoding'};
    }

    if (exists ($tempcfg{ 'output_encoding' }) ) {
	$config{ 'output_encoding' } = $tempcfg{ 'output_encoding' };
    }
    
    if (exists ($tempcfg{ 'input_encoding' }) ) {
	$config{ 'input_encoding' } = $tempcfg{ 'input_encoding' };
    }

    # print Dumper %config ; exit;
    return 1;
}

__DATA__
/* XPM */
static char *logo_data_32x32_16[] = {
/* columns rows colors chars-per-pixel */
"32 32 16 1",
"  c #251912",
". c #113C59",
"X c #224E6C",
"o c #53331C",
"O c #674B30",
"+ c #5D6162",
"@ c #306081",
"# c #547287",
"$ c #71899A",
"% c #885A33",
"& c #927456",
"* c #AD8D67",
"= c #9BA2A7",
"- c #B5B8B6",
"; c #D0D2D3",
": c #FEFEFE",
/* pixels */
";======$===$===-;;;:::::::::::::",
";=$X@@@@X.XXX@XX@XX@$:::::::::::",
":::;@X@X.#--;;=$@X@XXX=:::::::::",
"::::@@X.=::::::::;@X..X#::::::::",
"::::$@@.$::::::::::$X..X+:::::::",
"::::$@@X=:::::::::::=X..X#::::::",
"::::=@@.=::::::::::::$XXXX-:::::",
"::::=#@.=:::::::::::::#XXXX:::::",
"::::=#@X=:::::::::::::-@+XX-::::",
"::::-##X=::::::::::::::#X#X+::::",
"::::-##X=::::::::::::::=##X+;:::",
"::::-##X=::::::::::::::;$##+=:::",
"::::-$#+=:::::::::::::::$##++:::",
"::::-=$+=:::::::::::::::=#=+#:::",
"::::;=$+-:::::::::::::::==$+=;::",
"::::;==#=:::::::::::::::-===+;::",
"::::---+-:::::::::::::::;==+&;::",
"::::=O&&-:::::::::::::::;=+oo;::",
"::::-%Oo=:::::::::::::::=O%OO:::",
"::::-**O=:::::::::::::::&&*OO:::",
"::::-**%=:::::::::::::::***O*:::",
"::::***%*::::::::::::::;**OO=:::",
"::::-*&O=::::::::::::::;**OO;:::",
"::::*&%O-::::::::::::::**&O+::::",
"::::*%%O*:::::::::::::;&*oO;::::",
"::::=%OO=:::::::::::::&&Oo+:::::",
"::::*OOo=;;;;;;;;::;;*&OoO;:::::",
"::::&oOo   .;;;:;;;-*%oooO++=-::",
"::::&ooo   =:::::::&oooo;;     ;",
"::::ooo   =::::::=oooo&::-.   X;",
"::;*oo  oo&*=&*%ooo O=-=+oo+=-::",
"-&%%ooo    ooooOOO&&===-;;;:::::"
};
__END__
